from enum import Enum

from peaqevcore.common.spotprice.const import NORDPOOL, ENERGIDATASERVICE


class SpotPriceType(Enum):
    NordPool = NORDPOOL
    EnergidataService = ENERGIDATASERVICE
    Auto = 'Auto'


from dataclasses import dataclass, field
from datetime import datetime, time, timedelta
from ...models.hourselection.hourselection_options import HourSelectionOptions


@dataclass
class ScheduleSession:
    hourselection_options: HourSelectionOptions | None
    remaining_charge: float = 0
    desired_charge: float = 0
    starttime: datetime = datetime.min
    departuretime: datetime = datetime.min
    _hours_price: dict[datetime, float] = field(default_factory=lambda: {})
    _hours_charge: dict[datetime, float] = field(default_factory=lambda: {})
    _nh: list = field(default_factory=lambda: [])
    _ch: dict = field(default_factory=lambda: {})
    _mock_dt: datetime | None = None
    _init_ok: bool = False
    override_settings: bool = False
    _tomorrow_valid: bool = False

    def update_remaining_charge(self, charged_amount: float):
        rem = max(self.desired_charge - charged_amount,0)
        self.remaining_charge = round(rem, 2)

    @property
    def hours_price(self):
        return self._hours_price

    @hours_price.setter
    def hours_price(self, prices: list):
        today_date = datetime.now() if self._mock_dt is None else self._mock_dt
        price_dict = {}
        for idx, price in enumerate(prices[0]):
            price_dict[datetime.combine(today_date.date(), time(idx, 0))] = price
        if prices[1] is not None:
            self._tomorrow_valid = True
            tomorrow_date = today_date.date() + timedelta(days=1)
            for idx, price in enumerate(prices[1]):
                price_dict[datetime.combine(tomorrow_date, time(idx, 0))] = price
        else:
            self._tomorrow_valid = False
        self._hours_price = self._filter_price_dict(
            price_dict, self.starttime, self.departuretime
        )

    @property
    def hours_charge(self) -> dict:
        if self._init_ok:
            return self._hours_charge
        return dict()

    @hours_charge.setter
    def hours_charge(self, val):
        self._hours_charge = val
        self._init_ok = True

    @property
    def non_hours(self) -> list[datetime]:
        self._make_hours()
        return self._nh

    @property
    def caution_hours(self) -> dict[datetime, float]:
        self._make_hours()
        return self._ch

    def _make_hours(self) -> None:
        ch = {}
        nh = []
        _timer = datetime.combine(self.starttime.date(), time(self.starttime.hour, 0))
        today_dt = datetime.now() if self._mock_dt is None else self._mock_dt
        today_dt = today_dt.replace(minute=0, second=0, microsecond=0)  # start of the current hour
        while _timer < self.departuretime:
            if _timer >= today_dt:
                if _timer not in self.hours_charge.keys():
                    if self._tomorrow_valid or _timer.hour > today_dt.hour:
                        nh.append(_timer)
                elif 0 < self.hours_charge[_timer] < 1:
                    if self._tomorrow_valid or _timer.hour > today_dt.hour:
                        ch[_timer] = self.hours_charge[_timer]
            _timer += timedelta(hours=1)
        self._nh = nh
        self._ch = ch

    def _filter_price_dict(self, price_dict: dict, starttime: datetime, departuretime: datetime) -> dict:
        starttime_hour = datetime(starttime.year, starttime.month, starttime.day, starttime.hour)
        ret = {
            key: value
            for (key, value) in price_dict.items()
            if starttime_hour <= key <= (departuretime - timedelta(seconds=5))
        }
        return ret


import logging
from abc import abstractmethod
from dataclasses import dataclass, field
from statistics import mean, stdev
from datetime import date, datetime
from typing import Callable

_LOGGER = logging.getLogger(__name__)


@dataclass
class ISpotPriceDTO:
    state: float = 0
    today: list = field(default_factory=lambda: [])
    tomorrow: list = field(default_factory=lambda: [])
    average: float = 0
    stdev: float = 0
    currency: str = ""
    price_in_cent: bool = False
    tomorrow_valid: bool = False
    affected_date: date = None

    async def set_model(self, ret):
        try:
            self.today = list(ret.attributes.get("today"))
        except Exception as e:
            _LOGGER.exception(
                f"Could not parse today's prices. Unsolveable error. {e}"
            )
            return
        self.tomorrow_valid = bool(ret.attributes.get("tomorrow_valid", False))
        _tomorrow = ret.attributes.get("tomorrow", [])
        if _tomorrow is not None:
            self.tomorrow = list(_tomorrow)
        else:
            self.tomorrow = []
        self.currency = str(ret.attributes.get("currency", ""))
        self.state = ret.state
        self.average = self._set_average(ret)
        self.stdev = stdev(self.today) if len(self.today) > 1 else 0
        self.price_in_cent = self._set_price_in_cent(ret)
        self.affected_date = self._set_affected_date(self.affected_date, ret)

    @abstractmethod
    def _set_price_in_cent(self, ret) -> bool:
        pass

    @abstractmethod
    def _set_average(self, ret) -> float:
        pass

    @abstractmethod
    def _set_affected_date(self, affected_date, ret) -> date:
        pass


@dataclass
class EnergiDataServiceDTO(ISpotPriceDTO):

    def _set_price_in_cent(self, ret) -> bool:
            return bool(ret.attributes.get("use_cent", False))

    def _set_average(self, ret) -> float:
        try:
            return round(mean(self.today), 2)
        except Exception as e:
            _LOGGER.exception(
                f"Could not parse today's prices from EnergiDataService. Unsolveable error. {e}"
            )
            return 0

    def _set_affected_date(self, affected_date, ret) -> date:
        return datetime.now().date()


@dataclass
class NordpoolDTO(ISpotPriceDTO):
    
    def _set_price_in_cent(self, ret) -> bool:
        return bool(ret.attributes.get("price_in_cent", False))

    def _set_average(self, ret) -> float:
        try:
            return float(str(ret.attributes.get("average", 0)))
        except Exception as e:
            _LOGGER.exception(
                f"Could not parse today's prices from Nordpool. Unsolveable error. {e}"
            )
            return 0

    def _set_affected_date(self, affected_date: date, ret) -> date:
        try:
            rawdata = ret.attributes.get("raw_today", [])
            if not len (rawdata):
                _LOGGER.warning("No rawdata found for today")
                return affected_date
            new_date = rawdata[0]['start'].date()
            if rawdata and affected_date is None or affected_date != new_date:
                _LOGGER.debug(f"successfully setting {new_date} as spotprice affected date")
                return new_date
            else:
                _LOGGER.warning(
                    f"Unable to produce affected date for today. the data I got to play with was: {new_date}")
                return affected_date
        except Exception as e:
            _LOGGER.exception("Unable to parse date from raw_today data. Settings today's date as the valid one.", e)
            return datetime.now().date()




from datetime import datetime

from .update_scheduler_dto import UpdateSchedulerDTO
from ..hoursselection_service_new.models.hour_price import HourPrice
from ..hoursselection_service_new.models.permittance_type import PermittanceType
from ...models.chargecontroller_states import ChargeControllerStates
from .scheduler_service import Scheduler


class SchedulerFacade(Scheduler):
    def __init__(self, options):
        super().__init__(options)
        self.schedule_created = False

    @property
    def schedules(self) -> dict:
        if not self.scheduler_active:
            return {}
        return {1: {
            'schedule_starttime': self.model.starttime.strftime('%Y-%m-%d %H:%M'),
            'departure_time': self.model.departuretime.strftime('%Y-%m-%d %H:%M'),
            'charge_amount': self.model.desired_charge,
            'override_settings': self.model.override_settings,
            'remaning_charge': self.model.remaining_charge
        }}

    async def async_create_schedule(
        self,
        charge_amount: float,
        departure_time: datetime,
        schedule_starttime: datetime,
        override_settings: bool = False,
    ):
        if not self.scheduler_active:
            await self.async_create(
                charge_amount, departure_time, schedule_starttime, override_settings
            )
        self.schedule_created = True

    async def async_update_facade(self, model: UpdateSchedulerDTO):
        if self.schedule_created:
            await self.async_update(model)
            await self.async_check_states(model.chargecontroller_state)

    async def async_cancel_facade(self):
        await self.async_cancel()
        self.schedule_created = False

    async def async_check_states(self, chargecontroller_state: ChargeControllerStates):
        if not self.scheduler_active and self.schedule_created:
            await self.async_cancel_facade()
        elif (
            chargecontroller_state is ChargeControllerStates.Done
        ):
            await self.async_cancel_facade()

    def combine_future_hours(self, future_hours: list[HourPrice]):
        if not self.active:
            return future_hours

        dep = self.model.departuretime
        start = self.model.starttime.replace(minute=0, second=0, microsecond=0)
        hours_charge = self.model.hours_charge

        for hour_price in future_hours:
            if start <= hour_price.dt < dep:
                hour_price.permittance_type = PermittanceType.Scheduler
                if hour_price.dt in hours_charge:
                    hour_price.permittance = hours_charge[hour_price.dt]
                else:
                    hour_price.permittance = 0
        return future_hours

    #todo: redo these to work as future-hours instead. preferably reuse logic from hourselectionservice
    @property
    def non_hours(self) -> list[datetime]:
        return self.model.non_hours

    @property
    def caution_hours(self) -> dict[datetime,float]:
        """dynamic caution hours"""
        return self.model.caution_hours



from datetime import datetime
import logging
from ...models.locale.enums.price_type import PriceType
from ...models.locale.price.models.tiered_price import TieredPrice
from ..locale.locale_price_helpers import get_observed_peak

from ...models.locale.price.locale_price import LocalePrice
from .querytypes.queryservice import QueryService
from ...models.locale.peaks_model import PeaksModel
from ...models.locale.enums.sum_types import SumTypes
from ...models.locale.enums.time_periods import TimePeriods
from ...models.locale.sumcounter import SumCounter
from ...models.locale.queryproperties import QueryProperties
from .ilocale_query import ILocaleQuery

_LOGGER = logging.getLogger(__name__)

class LocaleQuery(ILocaleQuery):
    def __init__(
        self,
        sum_type: SumTypes,
        time_calc: TimePeriods,
        cycle: TimePeriods,
        sum_counter: SumCounter | None = None,
    ) -> None:
        self._peaks: PeaksModel = PeaksModel({})
        self._props = QueryProperties(sum_type, time_calc, cycle)
        self._sum_counter: SumCounter | None = sum_counter
        self._observed_peak_value: float = 0
        self._charged_peak_value: float = 0
        self._mock_dt: datetime | None = None
        self._price: LocalePrice = LocalePrice()

    @property
    def price(self) -> LocalePrice:
        return self._price
    
    @price.setter
    def price(self, value: LocalePrice):
        self._price = value

    def set_mock_dt(self, val: datetime | None):
        self._mock_dt = val

    @property
    def dt(self) -> datetime:
        return self._mock_dt if self._mock_dt is not None else datetime.now()

    @property
    def peaks(self) -> PeaksModel:
        if self._peaks.is_dirty:
            self._sanitize_values()
        return self._peaks

    @property
    def sum_counter(self) -> SumCounter:
        if self._sum_counter is not None:
            return self._sum_counter
        return SumCounter()

    @property
    def charged_peak(self) -> float:
        if self._peaks.is_dirty:
            self._sanitize_values()
        ret = self._charged_peak_value
        try:
            return round(ret, 2)
        except TypeError as t:
            _LOGGER.warning(t)
            return 0

    @charged_peak.setter
    def charged_peak(self, val):
        self._charged_peak_value = val

    @property
    def observed_peak(self) -> float:
        if self._peaks.is_dirty:
            self._sanitize_values()
        ret = (
            self.charged_peak
            if self._props.sumtype is SumTypes.Max
            else self.get_currently_obeserved_peak(self.dt)
        )
        try:
            return round(ret, 2)
        except TypeError as t:
            _LOGGER.warning(t)
            return 0

    @observed_peak.setter
    def observed_peak(self, val):
        self._observed_peak_value = val

    def get_currently_obeserved_peak(self, timestamp: datetime = datetime.now()) -> float:
        """gets the currently observed peak value for non-max type models. If daily max, override if the max registered today is higher."""
        ret = self._observed_peak_value

        if self.price.price_type == PriceType.Tiered:
            tiers:list[TieredPrice] = [s for s in self.price._values if isinstance(s, TieredPrice)]
            ret = get_observed_peak(self.price.price_type, [v for k, v in self._peaks.p.items()], tiers)

        if self.sum_counter.groupby == TimePeriods.Daily:
            if timestamp.day in [k[0] for k in self._peaks.p.keys()]:
                #_LOGGER.debug(f"exists as :{[v for k, v in self._peaks.p.items() if k[0] == timestamp.day][0]}. observed is {ret}. {self._peaks.p}. timestamp.day: {timestamp.day}")
                print(
                    f"exists as :{[v for k, v in self._peaks.p.items() if k[0] == timestamp.day][0]}. observed is {ret}. {self._peaks.p}. timestamp.day: {timestamp.day}")
                return max([v for k, v in self._peaks.p.items() if k[0] == timestamp.day][0], ret)
        print("returning observed", self._observed_peak_value)
        return ret

    def _sanitize_values(self):
        countX = lambda arr, x: len([a for a in arr if a[0] == x])
        if self.sum_counter.groupby == TimePeriods.Daily:
            duplicates = {}
            for k in self._peaks.p.keys():
                if countX(self._peaks.p.keys(), k[0]) > 1:
                    duplicates[k] = self._peaks.p[k]
            if len(duplicates):
                minkey = min(duplicates, key=duplicates.get) #type: ignore
                self._peaks.p.pop(minkey)
        while len(self._peaks.p) > self.sum_counter.counter:
            self._peaks.remove_min()
        self._peaks.is_dirty = False
        if self._props.sumtype is SumTypes.Max:
            self.charged_peak = self._peaks.max_value
        elif self._props.sumtype is SumTypes.Avg:
            self.observed_peak = self._peaks.min_value
            self.charged_peak = self._peaks.value_avg

    async def async_reset(self) -> None:
        await self._peaks.async_reset()
        self._observed_peak_value = 0
        self._charged_peak_value = 0

    async def async_set_query_service(self, service: QueryService) -> None:
        self._props.queryservice = service

    async def async_try_update(self, new_val, timestamp: datetime | None = None):
        _timestamp = timestamp or datetime.now()

        if not await self._props.queryservice.async_should_register_peak(dt=_timestamp):
            return
        _dt = (_timestamp.day, _timestamp.hour)
        if len(self.peaks.p) == 0:
            """first addition for this month"""
            await self.peaks.async_add_kv_pair(_dt, new_val)
            await self._peaks.async_set_month(_timestamp.month)
        elif _timestamp.month != self._peaks.m:
            """new month, reset"""
            await self.async_reset_values(new_val, _timestamp)
        else:
            await self.async_set_update_for_groupby(new_val, _dt)
        if len(self.peaks.p) > self.sum_counter.counter:
            await self.peaks.async_remove_min()
        await self.async_update_peaks()

    async def async_import_from_service(self, input_dict: dict) -> dict:
        """already good to go"""
        decorated_dict = {"m": datetime.now().month, "p": input_dict}
        await self.peaks.async_set_init_dict(decorated_dict)
        #await self.async_update_peaks()
        return decorated_dict

    async def async_set_update_for_groupby(self, new_val, dt):
        if self.sum_counter.groupby in [TimePeriods.Daily, TimePeriods.UnSet]:
            _datekeys = [k for k in self.peaks.p.keys() if dt[0] in k]
            #_LOGGER.debug(f"datekeys: {_datekeys}")
            if len(_datekeys):
                if new_val > self.peaks.p.get(_datekeys[0]):
                    await self.peaks.async_pop_key(_datekeys[0])
            await self.peaks.async_add_kv_pair(dt, new_val)
        elif self.sum_counter.groupby == TimePeriods.Hourly:
            if dt in self._peaks.p.keys():
                if new_val > self.peaks.p.get(dt):
                    await self.peaks.async_add_kv_pair(dt, new_val)
            else:
                await self.peaks.async_add_kv_pair(dt, new_val)

    async def async_update_peaks(self):
        if self._props.sumtype is SumTypes.Max:
            self.charged_peak = self._peaks.max_value
        elif self._props.sumtype is SumTypes.Avg:
            self.observed_peak = self._peaks.min_value
            self.charged_peak = self._peaks.value_avg
            print("updating peaks", self._peaks.p, self.observed_peak, self.charged_peak)

    async def async_reset_values(self, new_val, dt=datetime.now()):
        await self._peaks.async_clear()
        await self.async_try_update(new_val, dt)

    async def async_sanitize_values(self):
        countX = lambda arr, x: len([a for a in arr if a[0] == x])
        if self.sum_counter.groupby == TimePeriods.Daily:
            duplicates = {}
            for k in self._peaks.p.keys():
                if countX(self._peaks.p.keys(), k[0]) > 1:
                    duplicates[k] = self._peaks.p[k]
            if len(duplicates):
                minkey = min(duplicates, key=duplicates.get) #type: ignore
                self._peaks.p.pop(minkey)
        while len(self._peaks.p) > self.sum_counter.counter:
            await self._peaks.async_remove_min()
        self._peaks.is_dirty = False
        await self.async_update_peaks()



import time
from statistics import mean

class Average:
    def __init__(self, max_age: int, max_samples: int, precision: int = 2):
        self._readings = []
        self._average = 0
        self._smooth_average = 0
        self._max_age = max_age
        self._max_samples = max_samples
        self._precision = precision
        self._latest_update = 0

    @property
    def average(self) -> float:
        self._set_average()
        return round(self._average, self._precision)
    
    @property
    def smooth_average(self) -> float:
        self._set_smooth_average()
        return round(self._smooth_average, self._precision)

    def _set_smooth_average(self):
        x=0.1
        i = 1
        moving_averages = []
        moving_averages.append(self._readings[0][1])
        while i < len(self._readings):
            window_average = round((x*self._readings[i][1])+(1-x)*moving_averages[-1], 2)
            moving_averages.append(window_average)
            i += 1
        self._smooth_average = moving_averages[-1]
        
    def _set_average(self):
        try:
            self._average = mean([x[1] for x in self._readings])
        except ZeroDivisionError:
            self._average = 0

    def _remove_from_list(self):
        """Removes overflowing number of samples and old samples from the list."""
        while len(self._readings) > self._max_samples:
            self._readings.pop(0)
        gen = (
            x for x in self._readings if time.time() - int(x[0]) > self._max_age
        )
        for i in gen:
            if len(self._readings) > 1:
                # Always keep one reading
                self._readings.remove(i)

    def add_reading(self, val: float):
        self._readings.append((int(time.time()), round(val, 3)))
        self._latest_update = time.time()
        self._remove_from_list()
        self._set_average()

#------------
# from random import randrange

# a = Average(300, 200)
# averages = []
# smooth_averages = []

# for i in range(3600):
#     val = randrange(250, 1000)
#     a.add_reading(val)
   
#     averages.append(a.average)
#     smooth_averages.append(a.smooth_average)
#     #time.sleep(0.01)


# import matplotlib.pyplot as plt
# plt.plot(averages)
# plt.plot(smooth_averages)
# #plt.plot(averages2)
# #plt.plot(averages3)
# #plt.plot(maxes)
# #plt.plot(mins)
# plt.show()


import logging
import math
from statistics import mean
import time
from datetime import datetime, timedelta

_LOGGER = logging.getLogger(__name__)

MIN_SAMPLES = 2

def dt_from_epoch(epoch: int) -> str:
    return time.strftime("%Y-%m-%d %H:%M:%S", time.localtime(epoch))

class Gradient:
    def __init__(
        self, 
        max_age: int, 
        max_samples: int, 
        precision: int = 2,
        ignore: int|None = None, 
        outlier: float|None = None,
        smoothing_average: int = 1
    ):
        self._init_time = time.time()
        self._samples = []
        self._gradient = 0
        self._max_age = max_age
        self._max_samples = max_samples
        self._latest_update = 0
        self._ignore = ignore
        self._outlier = outlier
        self._precision = precision
        self._smoothing_average = smoothing_average

    @property
    def trend(self) -> float:
        return self._calculate_trend(use_decay=True)
    
    def _calculate_trend(self, use_decay: bool = False) -> float:
        self.prepare_gradient()
        data = self._samples
        n = len(data)
        if n < 2:
            return 0
        try:
            sum_x = sum([item[0] / 3600 for item in data])
            sum_y = sum([item[1] for item in data])
            sum_xy = sum([item[1]*item[0] / 3600 for item in data])
            sum_x2 = sum([(item[0] / 3600)**2 for item in data])
            slope = (n*sum_xy - sum_x*sum_y) / (n*sum_x2 - sum_x**2)
            time_since_last_sample = time.time() - self._latest_update
            decay_factor = math.exp(-time_since_last_sample / (60 * 60))  # Decay over an hour

            if use_decay:
                slope *= decay_factor

            if self._smoothing_average > 1:
                smoothed_slope = mean([slope] + [data[i][1] - data[i-self._smoothing_average][1] for i in range(self._smoothing_average, n)])
                ret = round(smoothed_slope, self._precision)
            else:
                ret = round(slope, self._precision)

            return ret if self._precision > 0 else int(ret)
        except ZeroDivisionError as e:
            return 0

    @property
    def gradient(self) -> float:
        self.prepare_gradient()
        return round(self._gradient, self._precision)

    @property
    def gradient_raw(self) -> float:
        self.prepare_gradient()
        return self._gradient

    @property
    def samples(self) -> int:
        return len(self._samples)

    @property
    def samples_raw(self) -> list:
        if len(self._samples) == 0:
            return []
        return sorted(self._samples, key=lambda x: x[0], reverse=True)

    @samples_raw.setter
    def samples_raw(self, lst):
        self._samples.extend(lst)
        self.prepare_gradient()

    @property
    def oldest_sample(self) -> str:
        if len(self._samples) > 0:
            return dt_from_epoch(self._samples[0][0])
        return str(datetime.min)

    @property
    def newest_sample(self) -> str:
        if len(self._samples) > 0:
            return dt_from_epoch(self._samples[-1][0])
        return str(datetime.min)
    
    def peak(self, idx=0) -> float:
        if len(self._samples) > 0:
            return self._samples[idx][1]
        return 0
    
    @property
    def is_clean(self) -> bool:
        return all([time.time() - self._init_time > 300, self.samples > 1])

    def prepare_gradient(self) -> None:
        self._inject_interim_values()
        self._remove_from_list()
        self._set_gradient()

    def _set_gradient(self) -> None:
        values = self._samples
        if len(values) == 1:
            self._gradient = 0
        elif len(values) - 1 > 0:
            try:
                ss = sorted(values, key=lambda x: x[0])
                x = (ss[-1][1] - ss[0][1]) / ((time.time() - ss[0][0]) / 3600)
                self._gradient = x
            except ZeroDivisionError as e:
                _LOGGER.warning({e})
                self._gradient = 0

    def _inject_interim_values(self) -> None:
        if self._samples:
            last_sample = self._samples[-1]
            if time.time() - last_sample[0] > 60 * 60:
                self._samples.append((int(time.time() - 30 * 60), last_sample[1]))

    def add_reading(self, val: float, t: float = time.time()):
        if self._ignore is None or self._ignore < val:
            if self._outlier is not None and len(self._samples) > 1 and abs(mean([s[1] for s in self._samples]) - val) > self._outlier:
                print("ignoring value", abs(mean([s[1] for s in self._samples]) - val))
                return
            if (int(t), round(val,3)) not in self._samples:
                self._samples.append((int(t), round(val, 3)))
                self._latest_update = t
                self._remove_from_list()
                self.prepare_gradient()

    def _remove_from_list(self):
        """Removes overflowing number of samples and old samples from the list."""
        while len(self._samples) > self._max_samples:
            self._samples.pop(0)
        old_samples = (x for x in self._samples if time.time() - int(x[0]) > self._max_age)
        for i in old_samples:
            if len(self._samples) > MIN_SAMPLES:
                # Always keep two readings to be able to calc trend
                self._samples.remove(i)

    async def async_set_gradient(self):
       self.prepare_gradient()

    async def async_add_reading(self, val: float, t: float = time.time()):
        self.add_reading(val, t)

    async def async_remove_from_list(self):
        self._remove_from_list()

    def predicted_time_at_value(self, target_value: float) -> datetime|None:
        if self._gradient is None or len(self._samples) < 2:
            return None
        current_gradient = self.trend
        if current_gradient == 0 or all([
            target_value < 0,
            self._samples[-1][1] > 0,
            current_gradient > 0
            ]):
            return None
        time_diff = timedelta(hours=(target_value - self._samples[-1][1]) / current_gradient)
        if time_diff < timedelta(0):
            return None
        return datetime.now() + time_diff
        
    def predicted_value_at_time(self, target_time: datetime) -> float|None:
        if self._gradient is None or target_time < datetime.now() or len(self._samples) < 2:
            return None
        current_gradient = self._gradient
        if current_gradient == 0:
            return None
        time_diff = target_time - datetime.now()
        expected_value = self._samples[-1][1] + (current_gradient * time_diff.total_seconds() / 3600)
        return round(expected_value,self._precision)





# ttt = [
#     - - 1701543099,
#   - 19.671,
# - - 1701542878,
#   - 19.7,
# - - 1701542747,
#   - 19.686,
# - - 1701542222,
#   - 19.7,
# - - 1701542215,
#   - 19.714,
# - - 1701542172,
#   - 19.729,
# - - 1701541371,
#   - 19.743,
# - - 1701541280,
#   - 19.729,
# - - 1701541181,
#   - 19.743,
# - - 1701539524,
#   - 19.757,
# - - 1701539478,
#   - 19.771,
# - - 1701539458,
#   - 19.786,
# - - 1701539338,
#   - 19.8,
# - - 1701539110,
#   - 19.814,
# - - 1701538982,
#   - 19.829,
# - - 1701538025,
#   - 19.814,
# - - 1701537533,
#   - 19.8,
# - - 1701537525,
#   - 19.814,
# - - 1701537263,
#   - 19.829,
# - - 1701537085,
#   - 19.857,
# - - 1701536982,
#   - 19.871,
# - - 1701536658,
#   - 19.886,
# - - 1701536226,
#   - 20.25,
# - - 1701536226,
#   - 20.2,
# - - 1701536226,
#   - 20.175,
# - - 1701536226,
#   - 20.06,
# - - 1701536226,
#   - 19.95,
# - - 1701536226,
#   - 19.9,
# - - 1701536019,
#   - 19.9,
# - - 1701536004,
#   - 19.929,
# - - 1701535978,
#   - 19.957,
# - - 1701535944,
#   - 19.971,
# ]

# ret = {}

# for idx, i in enumerate(ttt):
#     if idx % 2 == 0:
#         tt = i
#         ret[idx] = [tt]
#     else:
#         ret[idx-1].append(-float(i))

# tt = Gradient(max_age=7200, max_samples=100, precision=1)
# tt2 = Gradient(max_age=7200, max_samples=100, precision=1, smoothing_average=6)

# tt_values = []
# tt2_values = []
# x_values = []

# t = time.time()-3599
# tt.add_reading(54,t)
# tt2.add_reading(54,t)
# tt_values.append(tt.trend)
# tt2_values.append(tt2.trend)
# x_values.append(t)

# t = time.time()-3479
# tt.add_reading(53.8,t)
# tt2.add_reading(53.8,t)
# tt_values.append(tt.trend)
# tt2_values.append(tt2.trend)
# x_values.append(t)

# t = time.time()-3359
# tt.add_reading(53.8,t)
# tt2.add_reading(53.8,t)
# tt_values.append(tt.trend)
# tt2_values.append(tt2.trend)
# x_values.append(t)

# t = time.time()-3260
# tt.add_reading(53.1,t)
# tt2.add_reading(53.1,t)
# tt_values.append(tt.trend)
# tt2_values.append(tt2.trend)
# x_values.append(t)

# t = time.time()-3140
# tt.add_reading(53,t)
# tt2.add_reading(53,t)
# tt_values.append(tt.trend)
# tt2_values.append(tt2.trend)
# x_values.append(t)

# t = time.time()-3020
# tt.add_reading(53,t)
# tt2.add_reading(53,t)
# tt_values.append(tt.trend)
# tt2_values.append(tt2.trend)
# x_values.append(t)

# t = time.time()-2700
# tt.add_reading(52,t)
# tt2.add_reading(52,t)
# tt_values.append(tt.trend)
# tt2_values.append(tt2.trend)
# x_values.append(t)

# import matplotlib.pyplot as plt

# # Convert the timestamps to a more readable format
# x_values_readable = [datetime.fromtimestamp(x) for x in x_values]

# plt.figure(figsize=(10, 6))

# # Plot tt_values
# plt.plot(x_values_readable, tt_values, label='tt_values')

# # Plot tt2_values
# plt.plot(x_values_readable, tt2_values, label='tt2_values')

# plt.xlabel('Time')
# plt.ylabel('Values')
# plt.title('tt_values and tt2_values over time')
# plt.legend()

# #plt.show()

# print(tt_values)
# print(tt2_values)


from enum import Enum


class ObserverTypes(Enum):
    Test = "test"
    HubInitialized = "hub initialized" #1
    SpotpriceInitialized = "spotprice initialized"

    PowerCanaryDead = "power canary dead"
    PowerCanaryWarning = "power canary warning"

    PricesChanged = "prices changed"
    MonthlyAveragePriceChanged = "monthly average price changed"
    AdjustedAveragePriceChanged = "adjusted average price changed"
    DailyAveragePriceChanged = "daily average price changed"
    DynamicMaxPriceChanged = "dynamic max price changed"

    CarConnected = "car connected"
    CarDisconnected = "car disconnected" #can this be combined with CarConnected and a bool clause instead?
    CarDone = "car done"

    UpdateChargerEnabled = "update charger enabled"
    UpdateChargerDone ="update charger done" #is this the same as CarDone?
    UpdateLatestChargerStart = "update latest charger start"

    AuxStopChanged = "aux stop changed"
    MaxMinLimiterChanged = "max min limiter changed"
    TimerActivated = "timer activated"
    SchedulerCreated = "scheduler created"
    SchedulerCancelled = "scheduler cancelled"
    KillswitchDead = "killswitch dead"

    ProcessChargeController = "process charge controller"
    ProcessCharger = "process charger"

    ResetMaxMinChargeSensor = "reset max min charge sensor"
    UpdatePeak = "update peak"

    PrognosisChanged = "prognosis changed"
    UpdateOperation = "update operation"
    OffsetsChanged = "offsets changed"
    OffsetRecalculation = "offset recalculation"
    WatertempChange ="watertemp change"
    HvacToleranceChanged ="hvac tolerance changed"
    HvacPresetChanged ="hvac preset changed"
    TemperatureOutdoorsChanged="temperature outdoors changed"
    SetTemperatureChanged ="set temperature changed"

    GetPrices = "callback get prices"


import logging

from peaqevcore.common.spotprice.const import *
from peaqevcore.common.spotprice.energidataservice import EnergiDataServiceUpdater
from peaqevcore.common.spotprice.models.spotprice_type import SpotPriceType
from peaqevcore.common.spotprice.nordpool import NordPoolUpdater
from peaqevcore.common.spotprice.spotpricebase import SpotPriceBase
from peaqevcore.common.models.peaq_system import PeaqSystem

_LOGGER = logging.getLogger(__name__)


class SpotPriceFactory:

    sources = {
        SpotPriceType.NordPool: NordPoolUpdater,
        SpotPriceType.EnergidataService: EnergiDataServiceUpdater
    }

    @staticmethod
    def create(
        hub, 
        observer, 
        system: PeaqSystem,
        test:bool = False, 
        is_active: bool = False,
        custom_sensor: str = None,
        spotprice_type: SpotPriceType = SpotPriceType.NordPool
        ) -> SpotPriceBase:
        if test:
            return NordPoolUpdater(hub, test, system, observer)
        source = SpotPriceFactory.test_connections(hub.state_machine, spotprice_type, custom_sensor)
        return SpotPriceFactory.sources[source](hub, observer, system, test, is_active, custom_sensor)

    @staticmethod
    def test_connections(hass, spotprice_type: SpotPriceType, custom_sensor: str = None) -> SpotPriceType:
        if spotprice_type == SpotPriceType.NordPool:
            return SpotPriceType.NordPool
        sensor = hass.states.get(ENERGIDATASERVICE_SENSOR)       
        if sensor or (custom_sensor and custom_sensor == ENERGIDATASERVICE_SENSOR):
            return SpotPriceType.EnergidataService
        return SpotPriceType.NordPool
                

    


import logging
from datetime import datetime
from statistics import mean
from typing import Tuple

from peaqevcore.common.spotprice.models.average_type import AverageType
from peaqevcore.common.spotprice.models.dynamic_top_price_model import DynamicTopPriceModel

_LOGGER = logging.getLogger(__name__)


class DynamicTopPrice:
    def __init__(self):
        self.model = DynamicTopPriceModel()
        self.filterlen = {AverageType.SEVEN: 7, AverageType.THREE: 2}

    async def async_get_max(self, prices: list) -> Tuple[float, AverageType]:
        if not len(prices):
            _LOGGER.warning("No prices to calculate max from.")
            return 0, AverageType.ERROR
        try:
            await self.async_set_lists(prices)
            a = self.model.month[-1]
            b = self.model.three[-1]
            c = self.model.seven[-1]
            d = self.model.thirty[-1]
            rets = {a: AverageType.MONTH}
            rets.update(await self.async_measure_type(a, b, AverageType.THREE))
            rets.update(await self.async_measure_type(a, c, AverageType.SEVEN))
            rets.update(await self.async_measure_type(a, d, AverageType.THIRTY))
            return round(max(rets.keys()), 2), rets[max(rets.keys())]
        except Exception as e:
            _LOGGER.debug(f"Error in async_get_max: {e}")
            return 0, AverageType.ERROR

    async def async_set_lists(self, prices: list) -> None:
        self.model.three = await self.async_get_rolling(prices, 3)
        self.model.seven = await self.async_get_rolling(prices, 7)
        self.model.thirty = [mean(prices)]
        self.model.month = await self.async_get_current_month(prices)

    async def async_measure_type(self, month, measure, avg_type: AverageType) -> dict:
        ret: dict[float, AverageType] = {-1: avg_type}
        try:
            if measure > month:
                if avg_type == AverageType.THIRTY:
                    return {measure: avg_type}
                elif datetime.now().day >= self.filterlen[avg_type]:
                    match avg_type:
                        case AverageType.THREE:
                            gradient_measure = self.model.three
                        case AverageType.SEVEN:
                            gradient_measure = self.model.seven
                        case _:
                            return {}
                    gr = await self.async_set_gradient(gradient_measure)
                    if gr[-1] > 0:
                        ret = {measure: avg_type}
        except Exception as e:
            _LOGGER.debug(
                f"Error in async_measure_type: {e}. {avg_type}, {measure}, {month}"
            )
        return ret

    @staticmethod
    async def async_get_rolling(prices: list = [], days: int = 1):
        """get rolling seven day average from prices for each day."""
        rolling = []
        if not len(prices) or len(prices) < days:
            return []
        try:
            listlen = len(prices)
            for i in range((listlen - datetime.now().day), listlen):
                divident = prices[i - days : i]
                if not len(divident):
                    continue
                rolling.append(mean(divident))
        except Exception as e:
            _LOGGER.debug(f"Error in async_get_rolling: {e}")
        return rolling

    @staticmethod
    async def async_get_current_month(prices: list = []):
        rolling = []
        if not len(prices):
            return []
        try:
            listlen = len(prices)
            start_at = listlen - datetime.now().day
            for i in range(start_at, listlen):
                divident = prices[start_at : i + 1]
                if not len(divident):
                    break
                rolling.append(mean(divident))
        except Exception as e:
            _LOGGER.debug(f"Error in async_get_current_month: {e}")
        return rolling

    @staticmethod
    async def async_set_gradient(averages: list):
        """calculate the gradient of a list."""
        gradient = [0]
        for i in range(1, len(averages)):
            gradient.append(averages[i] - averages[i - 1])
        return gradient


# import asyncio

# prfices = [
#     1.49,
#     1.67,
#     1.58,
#     1.36,
#     0.85,
#     0.82,
#     0.79,
#     0.47,
#     0.49,
#     0.61,
#     0.31,
#     0.52,
#     0.64,
#     1.08,
#     1.48,
#     1.26,
#     0.85,
#     0.78,
#     0.89,
#     0.56,
#     0.57,
#     1,
#     0.6,
#     0.86,
#     1.53,
#     1.49,
#     0.97,
#     0.4,
#     0.78,
#     1.28,
#     0.8,
# ]


# async def run():
#     h = DynamicTopPrice()
#     gg = await h.async_get_max(prfices)
#     print(gg)


# asyncio.run(run())



from dataclasses import dataclass, field

from peaqevcore.models.chargecontroller_states import ChargeControllerStates


@dataclass
class UpdateSchedulerDTO:
    moving_avg24: float
    peak: float
    chargecontroller_state: ChargeControllerStates
    charged_amount: float = 0.0
    prices: list|None = field(default_factory=lambda: None)
    prices_tomorrow: list|None = field(default_factory=lambda: None)
    charge_per_hour: float = field(init=False)

    def __post_init__(self):
        try:
            self.charge_per_hour = self.peak - (self.moving_avg24 / 1000)
        except Exception as e:
            self.charge_per_hour = 0.0


import asyncio
import logging

import homeassistant.helpers.template as template #todo: fix this

from peaqevcore.common.models.observer_types import ObserverTypes
from peaqevcore.common.spotprice.const import NORDPOOL
from peaqevcore.common.spotprice.models.spotprice_dto import NordpoolDTO
from peaqevcore.common.spotprice.spotpricebase import SpotPriceBase
from peaqevcore.common.models.peaq_system import PeaqSystem

_LOGGER = logging.getLogger(__name__)


class NordPoolUpdater(SpotPriceBase):
    def __init__(self, hub, observer, system: PeaqSystem, test:bool = False, is_active: bool = True, custom_sensor: str = None):
        super().__init__(
            hub=hub,
            source=NORDPOOL,
            system=system,
            test=test,
            is_active=is_active,
            observer=observer,
            custom_sensor=custom_sensor
        )

    async def async_set_dto(self, ret, initial: bool = False) -> None:
        _result = NordpoolDTO()
        await _result.set_model(ret)
        if await self.async_update_set_prices(_result):
            if initial:
                await self.hub.async_update_prices(
                    [self.model.prices, self.model.prices_tomorrow]
                )
                _LOGGER.info("Nordpool service has been successfully.")
                _LOGGER.debug("Nordpool service has been initialized. Broadcasting...")
                await self.hub.observer.async_broadcast(ObserverTypes.SpotpriceInitialized)
            else:
                await self.hub.observer.async_broadcast(ObserverTypes.PricesChanged,[self.model.prices, self.model.prices_tomorrow])
            self._is_initialized = True

    def setup(self):
        try:
            entities = template.integration_entities(self.state_machine, NORDPOOL)
            _LOGGER.debug(f"Found {list(entities)} Spotprice entities for {self.model.source}.")
            if len(list(entities)) < 1:
                if hasattr(self.hub.options, "price"):
                    self.hub.options.price.price_aware = False  # todo: composition
                _LOGGER.error(
                    f"There were no Spotprice-entities. Cannot continue. with price-awareness."
                )
            if len(list(entities)) == 1 or self.custom_sensor:
                if self.custom_sensor:
                    _LOGGER.info(f"Using custom sensor for spot-prices: {self.custom_sensor}")
                self._setup_set_entity(self.custom_sensor if self.custom_sensor else list(entities)[0])
            else:
                _found: bool = False
                for e in list(entities):
                    if self._test_sensor(e):
                        _found = True
                        self._setup_set_entity(e)
                        break
                if not _found:
                    self.hub.options.price.price_aware = False  # todo: composition todo: fix this
                    _LOGGER.error(f"more than one Spotprice entity found. Cannot continue with price-awareness.")
        except Exception as e:
            if hasattr(self.hub.options, "price"):
                self.hub.options.price.price_aware = False  # todo: composition
            _LOGGER.error(f"I was unable to get a Spotprice-entity. Cannot continue with price-awareness: {e}")

    def _setup_set_entity(self, entity: str) -> None:
        self.model.entity = entity
        _LOGGER.debug(
            f"Nordpool has been set up and is ready to be used with {self.entity}"
        )
        asyncio.run_coroutine_threadsafe(
            self.async_update_spotprice(initial=True),
            self.state_machine.loop,
        )

    def _test_sensor(self, sensor: str) -> bool:
        """
        Testing whether the sensor has a set value for additional_costs_current_hour.
        This is the only way we can differ when there are multiple sensors present.
        """
        _LOGGER.debug(f"testing sensor {sensor}")
        state = self.state_machine.states.get(sensor)
        if state:
            _LOGGER.debug(f"got state")
            attr = state.attributes.get("additional_costs_current_hour", 0)
            _LOGGER.debug(f"testing attribute {attr}")
            if attr != 0:
                return True
        return False



from dataclasses import dataclass, field
from datetime import datetime
from abc import abstractmethod
from ..const import NON_HOUR, CAUTION_HOUR, CHARGING_PERMITTED
from ...hoursselection_service_new.models.hour_price import HourPrice
from ...timer.timer import Timer
from ...scheduler.scheduler_facade import SchedulerFacade
from typing import Tuple


class Hours:
    timer: Timer
    scheduler: SchedulerFacade

    def __init__(
        self, price_aware: bool, non_hours: list = [], caution_hours: list = []
    ):
        self._non_hours = non_hours
        self._caution_hours = caution_hours
        self._price_aware = price_aware
        self._is_initialized = False

    @property
    def state(self) -> str:
        if datetime.now().hour in self.non_hours:
            return NON_HOUR
        if datetime.now().hour in self.caution_hours:
            return CAUTION_HOUR
        return CHARGING_PERMITTED

    @property
    def price_aware(self) -> bool:
        return self._price_aware

    @property
    @abstractmethod
    def non_hours(self):
        pass

    @property
    @abstractmethod
    def is_initialized(self):
        pass

    @non_hours.setter
    @abstractmethod
    def non_hours(self, val):
        pass

    @property
    @abstractmethod
    def caution_hours(self):
        pass

    @caution_hours.setter
    @abstractmethod
    def caution_hours(self, val):
        pass

    @abstractmethod
    async def async_update_top_price(self, dyn_top_price) -> None:
        pass

    @property
    @abstractmethod
    def dynamic_caution_hours(self) -> dict:
        pass

    @property
    @abstractmethod
    def future_hours(self) -> list[HourPrice]:
        pass

    @abstractmethod
    def update_prices(self, prices: list = [], prices_tomorrow: list = []) -> None:
        pass

    @property
    @abstractmethod
    def options(self):
        pass

    @property
    @abstractmethod
    def offsets(self) -> dict:
        pass

    @property
    def prices(self) -> list:
        return []

    @prices.setter
    @abstractmethod
    def prices(self, val):
        pass

    @property
    def prices_tomorrow(self) -> list:
        return []

    @prices_tomorrow.setter
    @abstractmethod
    def prices_tomorrow(self, val):
        pass

    @property
    @abstractmethod
    def absolute_top_price(self):
        pass

    @property
    @abstractmethod
    def min_price(self):
        pass

    @property
    @abstractmethod
    def cautionhour_type_string(self) -> str:
        pass

    @property
    @abstractmethod
    def adjusted_average(self) -> float:
        pass

    @adjusted_average.setter
    @abstractmethod
    def adjusted_average(self, val):
        pass

    @abstractmethod
    async def async_get_average_kwh_price(self) -> Tuple[float | None, float | None]:
        pass

    @abstractmethod
    async def async_get_total_charge(self) -> Tuple[float, float | None]:
        pass

    @abstractmethod
    async def async_update_adjusted_average(self, val):
        pass

    @abstractmethod
    async def async_update_max_min(
        self,
        max_charge: float,
        session_energy: float | None = None,
        car_connected: bool = False,
        limiter: float = 0.0
    ):
        pass
